/* $Id:  $
 * Functions for controlling a st7565r connected graphical display over
 * SPI. (e.g. EA DOGL128)
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>
#include "lcdst7565r.h"
#if 0  /* For debugging */
#include "console.h"
#endif

uint8_t lcd_displaybuffer[1024];
uint8_t lcd_outputmode;

/* That file can be generated with fontconvert.pl from myfont.txt */
#include "lcd_defaultfont.h"

void lcd_setpixel(uint8_t x, uint8_t y, uint8_t c) {
  uint8_t page = y / 8;
  if ((x >= 128) || (y >= 64)) { /* Outside of screen */
    return;
  }
  if (c) {
    lcd_displaybuffer[page * 128 + x] |= _BV(y % 8);
  } else {
    lcd_displaybuffer[page * 128 + x] &= (uint8_t)~_BV(y % 8);
  }
  if (lcd_outputmode) {
    LCDPORT &= (uint8_t)~_BV(LCDA0PIN);
    /* Select page */
    lcd_sendspibyte(0xB0 | page);
    /* Jump to right column */
    lcd_sendspibyte(0x10 | ((x & 0xF0) >> 4));
    lcd_sendspibyte(0x00 | ((x & 0x0F) >> 0));
    LCDPORT |= _BV(LCDA0PIN);
    /* Write data byte */
    lcd_sendspibyte(lcd_displaybuffer[page * 128 + x]);
  }
}

/* draws a line with the Bresenham algorithm.
 * Mainly taken from wikipedia example. */
void lcd_drawline(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint8_t c) {
  int8_t Dx = x2 - x1;
  int8_t Dy = y2 - y1;
  uint8_t steep = (abs(Dy) >= abs(Dx));
  int8_t xstep = 1;
  int8_t ystep = 1;
  uint8_t x, y;
  uint8_t xDraw, yDraw;	
  int E;
  if (steep) {
    uint8_t swap;
    swap = x1;
    x1 = y1;
    y1 = swap;
    swap = x2;
    x2 = y2;
    y2 = swap;
    /* recompute Dx, Dy after swap */
    Dx = x2 - x1;
    Dy = y2 - y1;
  }
  if (Dx < 0) {
    xstep = -1;
    Dx = -Dx;
  }
  if (Dy < 0) {
    ystep = -1;
    Dy = -Dy;
  }
  E = 2 * Dy - Dx;
  y = y1;
  for (x = x1; x != x2; x += xstep) {
    if (steep) {
      xDraw = y;
      yDraw = x;
    } else {
      xDraw = x;
      yDraw = y;
    }
    lcd_setpixel(xDraw, yDraw, c);
    if (E > 0) {
      E += 2 * Dy - 2 * Dx;
      y = y + ystep;
    } else {
      E += 2 * Dy;
    }
  }
}

uint8_t lcd_printchar(uint8_t x, uint8_t y, uint8_t drawempty, uint8_t c, uint8_t txt) {
  uint16_t startoffs, endoffs;
  startoffs = pgm_read_word(&lcd_defaultfont.offsets[txt - 1]);
  endoffs = pgm_read_word(&lcd_defaultfont.offsets[txt]);
  while (startoffs < endoffs) {
    uint8_t i;
    uint8_t nbif = pgm_read_byte(&lcd_defaultfont.data[startoffs]);
    for (i = 0; i < 8; i++) {
      if (nbif & 0x01) {
        lcd_setpixel(x, y+i, c);
      } else { /* Pixel not set in font */
        if (drawempty) {
          lcd_setpixel(x, y+i, ((c == 0) ? 1 : 0));
        }
      }
      nbif >>= 1;
    }
    startoffs++;
    x++;
  }
  return x;
}

uint8_t lcd_printtext(uint8_t x, uint8_t y, uint8_t drawempty, uint8_t c, uint8_t * txt) {
  while (*txt) {
    x = lcd_printchar(x, y, drawempty, c, *txt);
    if (drawempty) {
      uint8_t i;
      for (i = 0; i < lcd_defaultfont.height; i++) {
        lcd_setpixel(x, y+i, ((c == 0) ? 1 : 0));
      }
    }
    x++; /* Space of 1 pixel */
    txt++;
  }
  return x;
}

uint8_t lcd_printtext_P(uint8_t x, uint8_t y, uint8_t drawempty, uint8_t c, PGM_P txt) {
  uint8_t t;
  while ((t = pgm_read_byte(txt))) {
    x = lcd_printchar(x, y, drawempty, c, t);
    if (drawempty) {
      uint8_t i;
      for (i = 0; i < lcd_defaultfont.height; i++) {
        lcd_setpixel(x, y+i, ((c == 0) ? 1 : 0));
      }
    }
    x++; /* Space of 1 pixel */
    txt++;
  }
  return x;
}

void lcd_drawimage(uint8_t x, uint8_t y, uint8_t flags, PGM_P img) {
  uint8_t cx; uint8_t cy;
  uint8_t bitpos = 0;
  uint8_t curibyte = 0, curtbyte = 0; /* Image and transparency bytes */
  uint8_t xs = pgm_read_byte(img++);
  uint8_t ys = pgm_read_byte(img++);
  uint8_t ifl = pgm_read_byte(img++);
  for (cy = 0; cy < ys; cy++) {
    for (cx = 0; cx < xs; cx++) {
      if (bitpos == 0) { /* Read next byte(s) */
        curibyte = pgm_read_byte(img++);
        if ((ifl & 0x01) == 0x01) {
          curtbyte = pgm_read_byte(img++);
        }
      }
      if (((ifl & 0x01) == 0x01) && ((curtbyte & 0x01) == 0x01)) {
        /* Transparency - nothing to do unless requested explicitly */
        if ((flags & LCD_IMGFLAG_TRANSWHITE) == LCD_IMGFLAG_TRANSWHITE) {
          lcd_setpixel(x+cx, y+cy, 0);
        }
        if ((flags & LCD_IMGFLAG_TRANSBLACK) == LCD_IMGFLAG_TRANSBLACK) {
          lcd_setpixel(x+cx, y+cy, 1);
        }
      } else { /* Non transparent pixel */
        if ((flags & LCD_IMGFLAG_INVERT) == LCD_IMGFLAG_INVERT) {
          lcd_setpixel(x+cx, y+cy, ((curibyte & 0x01) == 0x01) ? 0 : 1);
        } else {
          lcd_setpixel(x+cx, y+cy, ((curibyte & 0x01) == 0x01) ? 1 : 0);
        }
      }
      curibyte >>= 1;
      curtbyte >>= 1;
      bitpos++;
      if (bitpos >= 8) { bitpos = 0; }
    }
  }
}

/* Forces an update of the whole display. This goes rather fast. */
void lcd_updatewholedisplay(void) {
  uint8_t page; uint8_t col;
  for (page = 0; page < 8; page++) {
    LCDPORT &= (uint8_t)~_BV(LCDA0PIN);
    /* Select page */
    lcd_sendspibyte(0xB0 | page);
    /* Jump to column 0 */
    lcd_sendspibyte(0x10);
    lcd_sendspibyte(0x00);
    LCDPORT |= _BV(LCDA0PIN);
    for (col = 0; col < 128; col++) {
      lcd_sendspibyte(lcd_displaybuffer[page * 128 + col]);
    }
  }
}

void lcd_sendspibyte(uint8_t v) {
  uint8_t i;
  for (i = 0; i < 8; i++) {
    /* First put data on the bus */
    if (v & 0x80) {
      LCDPORT |= _BV(LCDSIPIN);
    } else {
      LCDPORT &= (uint8_t)~_BV(LCDSIPIN);
    }
    //_delay_loop_1(100);
    LCDPORT |= _BV(LCDSCLPIN); /* pull clock line high. Slave will read data here! */
    //_delay_loop_1(100);
    LCDPORT &= (uint8_t)~_BV(LCDSCLPIN); /* Take back SCL. */
    //_delay_loop_1(100);
    v <<= 1;
  }
}

void lcd_init() {
  /* Set port to output where relevant*/
  LCDDDR |= _BV(LCDA0PIN);
  LCDDDR |= _BV(LCDSCLPIN);
  LCDDDR |= _BV(LCDSIPIN);
  LCDDDR |= _BV(LCDCSPIN);
  
  /* Deselect chip (CS is inverted!) */
  LCDPORT |= _BV(LCDCSPIN);
  
  _delay_ms(100.0); /* Give it time to init */
  
  /* Now select and go */
  LCDPORT &= (uint8_t)~_BV(LCDCSPIN);
  
  lcd_sendspibyte(0x40); /* Start line 0 */
  lcd_sendspibyte(0xA1); /* Reverse display */
  lcd_sendspibyte(0xC0); /* select common output 0 - 63 */
  lcd_sendspibyte(0xA6); /* Display normal not inverted */
  lcd_sendspibyte(0xA2); /* BIAS 1/9  Duty 1/65 */
  lcd_sendspibyte(0x2F); /* Booster, regulator and follower on */
  lcd_sendspibyte(0xF8); /* set internal booster 4x */
  lcd_sendspibyte(0x00); /* ...cont. */
  lcd_sendspibyte(0x27); /* Voltage regulator set */
  lcd_sendspibyte(0x81); /* contrast set */
  lcd_sendspibyte(0x10); /* ...cont. */
  lcd_sendspibyte(0xAC); /* Static indicator set 0 */
  lcd_sendspibyte(0x00); /* ...cont. */
  lcd_sendspibyte(0xAF); /* Display on */
  lcd_updatewholedisplay();
}
